# ==========================================================
# vikiwiki.rb -- VikiWiki-WEBrick Trial-Kit package.
#
# Copyright (C) 2005 Z80-OOLONG. (NAKATSUKA, Yukitaka)
# ==========================================================

require "fileutils"
require "find"
require "webrick"
require "base64"
require "thread"

require "vikiwiki/config"

# ----------------------------------------------------------

$:.unshift($VIKIWIKIDIR)
$:.unshift("#{$VIKIWIKIDIR}/lib")

require "vikiwikiweb"

if /^1\.8.*/ === $WIKIVERSION then
  require "webrickextend"
  require "erbextend"
elsif /^1\.9.*/ === $WIKIVERSION then
  require "extension/webrick"
  require "extension/erb"
end

# ----------------------------------------------------------

module VikiWiki
  ############
  # WEBrick servlet class.
  ############

  class Servlet < WEBrick::HTTPServlet::AbstractServlet
    ############
    # Utility module

    module Util
      ############
      # sanitize and expand path

      def expand_path(path)
        if (/[|&;`$<>]/ =~ path) then
          raise WEBrick::HTTPStatus::InternalServerError
        end
        path = File::expand_path(path.untaint); path.gsub!(%r|^/+|, "/")
        return path
      end
      module_function :expand_path

      ############
      # definition of constant

      BASEDIR    = expand_path("#{ENV['HOME']}/VikiWiki")
      BASEURI    = "/"
      SCRIPTNAME = "/vikiwiki"
      STATICDIR  = "#{BASEDIR}/static"
      STATICURI  = "/static"
    end

    include Util

    ############
    # class Config. This class is for VikiWiki configuration and
    # a subclass of Hash.

    class Config < Hash
      include Util

      ############
      # Default configuration. See also vikiwiki/parameter.txt in vikiwiki.zip

      DEFAULT_CONFIG = VikiWiki::CONF.clone
      DEFAULT_CONFIG["AUTHOR"]      = "Administrator"
      DEFAULT_CONFIG["ADMIN"]       = "admin"
      DEFAULT_CONFIG["ADDRESS"]     = "admin@example.com"
      DEFAULT_CONFIG["AUTHTYPE"]    = "none"
      DEFAULT_CONFIG["BASEDIR"]     = BASEDIR
      DEFAULT_CONFIG["BASEURI"]     = BASEURI
      DEFAULT_CONFIG["ENCODING"]    = $VIKIWIKIENCODE
      DEFAULT_CONFIG["SCRIPT_NAME"] = SCRIPTNAME
      DEFAULT_CONFIG["STATICDIR"]   = STATICDIR
      DEFAULT_CONFIG["STATICURI"]   = STATICURI
      DEFAULT_CONFIG["WIKIFARM"]    = false
      DEFAULT_CONFIG["TITLE"]       = "VikiWiki - page"
      DEFAULT_CONFIG["STATICTITLE"] = "alias"
      DEFAULT_CONFIG["AUTHREALM"]   = "VikiWiki User"

      Default = self.new().freeze

      def initialize(conf = {})
        self.clear; self.update(DEFAULT_CONFIG)
        conf = conf.dup

        if conf["BASEDIR"] then
          conf["BASEDIR"] = expand_path(conf["BASEDIR"])
          conf["STATICDIR"] ||= "#{conf['BASEDIR']}/static"
        end

        ["WIKISTYLES", "LANGS"].each do |key|
          if conf[key] then
            self[key].unshift(*conf[key]); conf.delete(key)
          end
        end

        if conf["REPLACE_WORDS"] then
          conf["REPLACE_WORDS"].each do |key, value|
            if Proc === value then
              conf["REPLACE_WORDS"][key] = Promise.new(value)
            end
          end
        end

        self.update(conf)
      end
    end

    ############
    # class Promise. an object for delay evaluation. This class is used
    # in @conf["REPLACE_WORDS"].

    class Promise
      def initialize(expr); @expr = expr; end
      def force; @expr.force; end
      def to_s; force.to_s; end
      def to_i; force.to_i; end

      class ::String
        def force; eval(self); end
        def delay; VikiWiki::Servlet::Promise.new(self); end
      end

      class ::Proc
        def force; self.call; end
        def delay; VikiWiki::Servlet::Promise.new(self); end
      end
    end

    ############
    # class UserDB. a passwd object for WEBrick::HTTPAuth::BasicAuth.

    class UserDB
      include WEBrick::HTTPAuth::UserDB
      KEYLEN = 16

      def initialize(etcdir)
        @path = "#{etcdir}/passwd"
        @auth_type = WEBrick::HTTPAuth::BasicAuth
        @passwd = Hash::new([]); @mtime = Time::at(0)
      end

      def reload
        if File::file?(@path) && \
           ((mtime = File::mtime(@path)) > @mtime) then
          File::foreach(@path) do |line|
            user, pass, salt = line.split(":")
            @passwd[user] = [pass, salt]
          end
          @mtime = mtime
        end
      end

      def get_passwd(realm, user, reload_db)
        reload_db && reload(); return @passwd[user][0]
      end

      def set_passwd(realm, user, pass)
        salt = WEBrick::Utils::random_string(KEYLEN)
        pass = pass.crypt(salt)
        @passwd[user] = [pass, salt]
        text = @passwd.inject("") do |text, (user, (pass, salt))|
          text << "#{user}:#{pass}:#{salt}\n"; text
        end
        File::fwrite(@path, text); return pass
      end
    end

    class << self
      Instances = Hash.new
      InstancesMutex = Mutex.new

      def mounted(dir, conf = Hash.new)
        # conf["SCRIPT_NAME"] must be mount point.
        conf = conf.dup; conf["SCRIPT_NAME"] = dir
        return dir, conf
      end

      def get_instance(server, conf)
        InstancesMutex.synchronize do
          Instances[conf["SCRIPT_NAME"]] ||= self.new(server, conf)
        end
      end
    end

    def initialize(server, conf)
      super
      config_environment(conf)
      check_repository
      check_mount(server)
      set_auth
    end

    def do_GET(request, response)
      (Thread.start do; $SAFE += 1
        @auth && @auth.authenticate(request, response)

        cgi = request.to_cgi
        $VIKIWIKIOUT = response.to_out

        VikiWiki::Web.new(VikiWiki::Sys.new("/dev/null", @conf, cgi))
      end).join
    end

    alias do_POST do_GET

    private

    ############
    # set VikiWiki configuration hash and reset environment variable.

    def config_environment(conf)
      # set @conf to VikiWiki configuration
      @conf = Config.new(conf)

      # add @conf["BASEDIR"]/{,lib} to Ruby include path.
      $:.include?(@conf["BASEDIR"]) || $:.unshift(@conf["BASEDIR"])
      $:.include?("#{@conf['BASEDIR']}/lib") \
                           || $:.unshift("#{@conf['BASEDIR']}/lib")

      # set environment variable TMPDIR, TMP, PATH.
      $VIKIWIKITMP = ENV['TMPDIR'] = ENV['TMP'] = "#{@conf['BASEDIR']}/tmp"
      ENV["PATH"] = "/usr/bin:/usr/sbin:/bin:/sbin"
    end

    ############
    # check VikiWiki repository.
    # If directory @conf["BASEDIR"] is not exist, a VikiWiki repository
    # is installed under @conf["BASEDIR"].

    def check_repository
      if !FileTest.directory?(@conf["BASEDIR"]) then
        FileUtils.mkdir_p(@conf["BASEDIR"])
        Dir.chdir("#{$VIKIWIKIDIR}/master") do
          Find::find(".") do |f|
            f.gsub!(%r|\./|, "")
            if FileTest.directory?(f) && !(/^\.\.?$/ =~ f) then
              FileUtils.mkdir_p("#{@conf["BASEDIR"]}/#{f}")
              FileUtils.chmod(0755, "#{@conf["BASEDIR"]}/#{f}")
            elsif FileTest.file?(f) then
              FileUtils.install(f, "#{@conf["BASEDIR"]}/#{File.dirname(f)}",
                                {:mode => 0644})
            end
          end
        end
      end
    end

    ############
    # check if @conf["BASEDIR"]/{theme,img,script} and @conf["STATICDIR"] is
    # mounted.

    def check_mount(server)
      check_mount_uri(server,
        @conf["THEMEURI"] || "#{@conf["BASEURI"]}/theme",
        @conf["THEMEDIR"] || "#{@conf["BASEDIR"]}/theme")

      check_mount_uri(server,
        @conf["IMGURI"] || "#{@conf["BASEURI"]}/img",
        @conf["IMGDIR"] || "#{@conf["BASEDIR"]}/img")

      check_mount_uri(server,
        "#{@conf["BASEURI"]}/script", "#{@conf["BASEDIR"]}/script")

      check_mount_uri(server, @conf["STATICURI"], @conf["STATICDIR"])
    end

    def check_mount_uri(server, uri, dir)
      uri = uri.gsub(%r|/+|, "/"); uri.gsub!(%r|/$|, "")
      dir = expand_path(dir)
      if !already_mounted_uri?(server, uri, dir) then
        server.mount(uri, WEBrick::HTTPServlet::FileHandler, dir)
      end
    end

    def already_mounted_uri?(server, uri, dir)
      servlet, options, _, _ = server.search_servlet(uri)
      (servlet == WEBrick::HTTPServlet::FileHandler) && (dir == options[0])
    end

    ############
    # set authentication.

    def set_auth
      @auth = nil
      if @conf["AUTHTYPE"] != "none" then
        @conf["AUTHTYPE"] = "www"

        realm = @conf["AUTHREALM"] || "VikiWiki User"
        etcdir = @conf["ETCDIR"] || "#{@conf["BASEDIR"]}/etc"
        userdb = UserDB::new(etcdir)

        if !userdb.get_passwd(realm, @conf["ADMIN"], true) then
          userdb.set_passwd(realm, @conf["ADMIN"], @conf["ADMIN"])
        end

        @auth = WEBrick::HTTPAuth::BasicAuth.new(
          :Realm  => realm, :UserDB => userdb
        )
      end
    end
  end

  ############
  # VikiWiki-WEBrick server class.

  class WEBrickServer
    def initialize(conf = {})
      @conf = conf; @server = nil

      port = (conf["PORT"] || 10080); conf.delete("PORT")
      scriptname = (conf["SCRIPT_NAME"] || VikiWiki::Servlet::SCRIPTNAME)

      @server = WEBrick::HTTPServer.new(:Port => port)
      @server.mount(scriptname, VikiWiki::Servlet, @conf)

      trap("INT"){ @server.shutdown }; trap("TERM"){ @server.shutdown }
      @server.start
    end

    ############
    # WEBrickServer::start == WEBrickServer::new
    class << self; alias start new; end
  end

  ############
  # VikiWiki::Config == VikiWiki::Servlet::Config
  Config = Servlet::Config
end

# ----------------------------------------------------------

VikiWikiWEBrick = VikiWiki::WEBrickServer

# ----------------------------------------------------------

($0 == __FILE__) && VikiWiki::WEBrickServer::start()

# ==========================================================
